/*
	Relay -- a tool to record and play Quake2 demos
	Copyright (C) 2000 Conor Davis

	This program is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 2
	of the License, or (at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

	Conor Davis
	cedavis@planetquake.com
*/

// g_local.h -- local definitions for game module
#ifndef __LOCAL_H
#define __LOCAL_H

#include <stdio.h>

// define GAME_INCLUDE so that game.h does not define the
// short, server-visible gclient_t and edict_t structures,
// because we define the full size ones in this file
#define GAME_INCLUDE

#include "shared.h"
#include "cmd.h"
#include "dm2.h"
#include "mem.h"
#include "menu.h"
#include "pak.h"

// the "gameversion" client command will print this plus compile date
#define GAMEVERSION "q2server"

#define FRAMETIME		0.1

// memory tags to allow dynamic memory to be cleaned up
#define TAG_GAME	765 	// clear when unloading the dll
#define TAG_LEVEL	766 	// clear when loading a new level

// handedness values
#define RIGHT_HANDED			0
#define LEFT_HANDED 			1
#define CENTER_HANDED			2

typedef struct
{
	gclient_t	*clients;		// [maxclients]

	// store latched cvars here that we want to get at often
	int 		maxclients;
	int 		maxentities;

	int 		relayflags;
	int 		player;
} game_locals_t;

typedef struct
{
	int 		framenum;
	float		time;
} level_locals_t;

typedef enum {
	F_INT, 
	F_FLOAT,
	F_LSTRING,			// string on disk, pointer in memory, TAG_LEVEL
	F_GSTRING,			// string on disk, pointer in memory, TAG_GAME
	F_VECTOR,
	F_ANGLEHACK,
	F_EDICT,			// index on disk, pointer in memory
	F_ITEM, 			// index on disk, pointer in memory
	F_CLIENT,			// index on disk, pointer in memory
	F_FUNCTION,
	F_MMOVE,
	F_IGNORE
} fieldtype_t;

typedef struct
{
	char	*name;
	int 	ofs;
	fieldtype_t type;
	int 	flags;
} field_t;

extern	field_t fields[];

extern	game_import_t	gi;
extern	game_export_t	globals;
extern	game_locals_t	game;
extern	level_locals_t	level;

extern	edict_t 		*g_edicts;

#define FOFS(x) (int)&(((edict_t *)0)->x)

#define random()	((rand () & 0x7fff) / ((float)0x7fff))
#define crandom()	(2.0 * (random() - 0.5))

extern	cvar_t	*deathmatch;
extern	cvar_t	*password;
extern	cvar_t	*spectator_password;
extern	cvar_t	*needpass;
extern	cvar_t	*maxclients;
extern	cvar_t	*maxspectators;
extern	cvar_t	*dedicated;

extern	cvar_t	*filterban;

extern	cvar_t	*run_pitch;

extern	cvar_t	*flood_msgs;
extern	cvar_t	*flood_persecond;
extern	cvar_t	*flood_waitdelay;

//============================================================================

struct gclient_s
{
	// known to server
	player_state_t	ps; 			// communicated by server to clients
	int 			ping;

	char			userinfo[MAX_INFO_STRING];
	char			netname[16];

	vec3_t			v_angle;		// remove?
	vec3_t			cmd_angles;
	int 			buttons;
	int 			oldbuttons;
	int 			latched_buttons;
	pmove_state_t	old_pmove;

	int 			relayflags;
	int 			player;
	float			dist;

	short			oldinventory[MAX_ITEMS];
	short			inventory[MAX_ITEMS];

	char			oldlayout[MAX_MSGLEN];
	char			layout[MAX_MSGLEN];
	qboolean		layout_modified;				// so we don't need to strcmp the old and new menu's
	qboolean		show_layout;

	float			bigupdate_time;

	menu_t			*curmenu;
};

struct edict_s
{
	entity_state_t	s;
	struct gclient_s	*client;	// NULL if not a player
									// the server expects the first part
									// of gclient_s to be a player_state_t
									// but the rest of it is opaque

	qboolean	inuse;
	int 		linkcount;

	// FIXME: move these fields to a server private sv_entity_t
	link_t		area;				// linked to a division node or leaf
	
	int 		num_clusters;		// if -1, use headnode instead
	int 		clusternums[MAX_ENT_CLUSTERS];
	int 		headnode;			// unused if num_clusters != -1
	int 		areanum, areanum2;

	//================================

	int 		svflags;
	vec3_t		mins, maxs;
	vec3_t		absmin, absmax, size;
	solid_t 	solid;
	int 		clipmask;
	edict_t 	*owner;


	// DO NOT MODIFY ANYTHING ABOVE THIS, THE SERVER
	// EXPECTS THE FIELDS IN THAT ORDER!

	//================================
	int 		movetype;
	int 		flags;

	char		*model;
	float		freetime;			// sv.time when the object was freed
	
	//
	// only used locally in game, not by server
	//
	char		*classname;
	vec3_t		velocity;
	int 		light_level;
};

extern game_import_t	real_gi;
extern dm2_t	dm2in;
extern short	edict_table[MAX_EDICTS];
extern byte 	current_connected[MAX_CLIENTS/8];
extern byte 	old_connected[MAX_CLIENTS/8];
extern PFILE	*infile;
extern vec3_t	spawnpoints;
extern int		numportals;
extern cvar_t	*demospeed;

// rp_dm2.c
extern void UpdateEntity(entity_state_t *es, qboolean active);
extern int PreFrame_Parse(block_t *block);
extern int Frame_Parse(block_t *block);

// rp_main.c
extern int AdvanceFrame();

// rp_svcmds.c
extern qboolean SV_FilterPacket (char *from);

// rp_view.c
extern void ClientBeginServerFrame (edict_t *ent);
extern void ClientEndServerFrame (edict_t *ent);

// rp_q2utils.c
extern edict_t *G_Find (edict_t *from, int fieldofs, char *match);
extern edict_t *findradius (edict_t *from, vec3_t org, float rad);
extern void G_SetMovedir (vec3_t angles, vec3_t movedir);
extern char *G_CopyString (char *in);
extern void G_InitEdict (edict_t *e);
extern edict_t *G_Spawn (void);
extern void G_FreeEdict (edict_t *ed);

// rp_utils.h
extern void UpdatePlayerOrigin(edict_t *ent);
extern int ChangePlayer(edict_t *ent, int index);

// rp_menus.h
enum
{
  MENU_BAD,
  MENU_MAIN,
  MENU_DEMO,
  MENU_PLAYERS,
  MENU_SETTINGS
};

extern void UpdateLayout(edict_t *ent);
extern void OpenMenu(edict_t *ent, void (*Show)(menu_t *));
extern void CloseMenu(edict_t *ent);
extern void CloseAllMenus(edict_t *ent);
extern void UpdateAllMenus(int id);
extern void Select_OpenMenu(menu_t *menu, menuitem_t *item, int key);
extern void DemoMenu_Show(menu_t *menu);
extern void PlayersMenu_Show(menu_t *menu);
extern void SettingsMenu_Show(menu_t *menu);
extern void MainMenu_Show(menu_t *menu);

#endif // __LOCAL_H
